#define PRINT 0
#include <stdio.h>
#include "m3.h"
/* 
 * Calcula a Homografia H4p[4x4] que leva um paralelograma no espaco (u,v) para 
 * um quadrilatero qualquer no espaco (x,y)
 */
int homography4p(double u[4], double v[4], double x[4], double y[4], double H4p[4*4])
{
    double N[3*3];          /* [N] transforms the rectangle given by (X,Y) in a unity square */
    double Ninv[3*3] = { u[1]-u[0], u[3]-u[0],     u[0],  
                         v[1]-v[0], v[3]-v[0],     v[0],   
                                 0,         0,      1};
    /* [M] transforms the deformed unit square into the {u,v) */
    double M[3*3] = { x[1]-x[0], x[3]-x[0],     x[0],  
                      y[1]-y[0], y[3]-y[0],     y[0],   
                               0,         0,      1};
    double P[3*3] = { 0,0,0,  0,0,0,  0,0,0};             /* initialize with zeros */
    double b[3]   = {x[2],y[2],1.};
    double xx[3];
    double Temp[3*3];
    double alpha,betha;
    double Minv[3*3]={0,0,0, 0,0,0, 0,0,0};
    double detM=m3Inv(M,Minv);
    double detN=m3Inv(Ninv,N);


    if (PRINT) { 
        printf("detN=%f detM=%f\n",detN,detM); 
        m3PrintMat("[N]",N); m3PrintMat("[Ninv]",Ninv);
        m3PrintMat("[M]",M); m3PrintMat("[Minv]",Minv); 
        m3PrintVet("{uv2}=",b); 
    }

    m3MultAb(Minv,b,xx);
    alpha=xx[0];
    betha=xx[1];
    if (PRINT) {
        printf("\n alpha=%f  e   betha=%f\n",alpha,betha);

        b[0]=xx[0]; b[1]=y[0]; b[2]=1.;
        m3MultAb(Minv,b,xx);
        m3PrintVet("{0}=",xx);

        b[0]=xx[1]; b[1]=y[1]; b[2]=1.;
        m3MultAb(Minv,b,xx);
        m3PrintVet("{1}=",xx);

        b[0]=xx[2]; b[1]=y[2]; b[2]=1.;
        m3MultAb(Minv,b,xx);
        m3PrintVet("{2}=",xx);

        b[0]=xx[3]; b[1]=y[3]; b[2]=1.;
        m3MultAb(Minv,b,xx);
        m3PrintVet("{3}=",xx);
    }

    P[0]=alpha; P[4]=betha; P[6]=1-betha; P[7]=1-alpha; P[8]=alpha+betha-1;
    if (PRINT) m3PrintMat("[P]",P);

    m3MultAB(M,P,Temp);
    m3MultAB(Temp,N,H4p);

    if (PRINT) m3PrintMat("[H4p]",H4p);

    return 0;
}