/* IUP/OpenGL image example

  Uses IUP for user interface
       IM for image I/O and capture
       OpenGL for drawing

  Needs IUP: comctl32.lib iup.lib iupcontrols.lib iupgl.lib cd.lib cdiup.lib 
        OpenGL: opengl32.lib glu32.lib

  Libraries Available at:
    http://www.tecgraf.puc-rio.br/iup/

  There are two canvases. 
  One for that uses the OpenGL process (left) and 
  one for you to disply your results (right). 
  
  Version:
            0.9 

  Authors:
            Marcelo Gattass
            Antonio Scuri

  Tecgraf/PUC-Rio - 2007
*/

#ifdef WIN32            
#include <windows.h>    /* necessary because of the Microsoft OpenGL headers dependency */
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <iup.h>
#include <iupcontrols.h>
#include <iupgl.h>

#include "image.h"
#include "homografia.h"
#include "m3.h"
#include "filter.h"

void ewa_filter(double* Hinv, float* rgb, double posx, double posy, double u0, double v0);

/* Global variables of the module */
struct _globals
{
 Image*  image;  /* current image */
 Ihandle *left_canvas, *right_canvas, *status_label, *dialog;  /* IUP handlers */
 int flags[4];   /* control flags */
 double value1, value2;  /* control values */
} glb ={ NULL, NULL, NULL, NULL, NULL, 1, 0, 0, 0, 0., 0. };

/* update application canvas and msg bar */

/*
 * Callbacks
 */

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int left_repaint_cb(Ihandle* self)
{
 int w = 256;
 int h = 256;

 IupGLMakeCurrent(self);        /* activates this GL Canvas as the current drawing area. */
 glClear(GL_COLOR_BUFFER_BIT);  /* clears the back buffer */

 if (glb.image!=NULL)
 {
  float* pBuffer = imgGetData(glb.image);
  w = imgGetWidth(glb.image);
  h = imgGetHeight(glb.image);
  
  /* Draws the captured image at (0,0) */
  glRasterPos2f(0.f, 0.f);
  glDrawPixels(w, h, GL_RGB, GL_FLOAT, pBuffer);
 }

 glColor3f(1.f,1.f,0.f);
 glBegin(GL_LINE_LOOP);
 glVertex2i(0,0);
 glVertex2i(w, 0);
 glVertex2i(w,h);
 glVertex2i(0,h);
 glEnd();


 IupGLSwapBuffers(self);        /* swap data from back buffer to front buffer */
 return IUP_DEFAULT;
} 

/* FACA ESTA FUNCAO CORRETAMENTE [opcional]*/
//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
void ogl_texture(double* Hinv)
{
 if (glb.image)
 {
  float plane[4];
  int w = imgGetWidth(glb.image);
  int h = imgGetHeight(glb.image);
  double delta0 = w*glb.value1;
  double delta1 = 0.5*w*glb.value2;

  /*temp[0] = -delta0;
  temp[1] = 0.0;
  temp[2] = 1.0;
  m3MultAb(Hinv, temp, pxIni);
  pxIni[0] /= pxIni[2];
  pxIni[1] /= pxIni[2];

  temp[0] = w+delta0;
  temp[1] = 0.0;
  temp[2] = 1.0;
  m3MultAb(Hinv, temp, pxFin);
  pxFin[0] /= pxFin[2];
  pxFin[1] /= pxFin[2];

  /*temp[0] = 0.0;
  temp[1] = 0.0;
  temp[2] = 1.0;
  m3MultAb(Hinv, temp, pyIni);
  pyIni[0] /= pyIni[2];
  pyIni[1] /= pyIni[2];

  temp[0] = h;
  temp[1] = 0.0;
  temp[2] = 1.0;
  m3MultAb(Hinv, temp, pyFin);
  pyFin[0] /= pyFin[2];
  pyFin[1] /= pyFin[2];*/

  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  
  plane[0] = - 1.0f / (-delta0 - (w+delta0));
  plane[1] = 0.0;
  plane[2] = 0.0;
  plane[3] = -delta0 / (-delta0 - (w+delta0));
  glTexGenfv(GL_S, GL_OBJECT_PLANE, plane);

  plane[0] = 0.0;
  plane[1] = - 1.0f / (-h);
  plane[2] = 0.0;
  plane[3] = 0.0;
  glTexGenfv(GL_T, GL_OBJECT_PLANE, plane);


  glEnable(GL_TEXTURE_GEN_S);
  glEnable(GL_TEXTURE_GEN_T);

  //glDisable(GL_BLEND);
  glBegin(GL_QUADS);
   glVertex2d(-delta0,0.f);
   glVertex2d(w+delta0,0);
   glVertex2d(w-delta1,h);
   glVertex2d(delta1,h);
  glEnd();

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);
 }
}

/* FACA ESTA FUNCAO CORRETAMENTE  */
//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
void my_texture( )
{
 if (glb.image)
 {
  float rgb[3];
  double posx, posy, posxIni, posxFin, incxIni, incxFin;
  double H[3*3], Hinv[3*3], Jxy[4], ux, vx, uy, vy, u0, v0;
  double u[4], v[4], x[4], y[4], p1[3], p2[3], tmp[3], du[2], dv[2];

  int w = imgGetWidth(glb.image);
  int h = imgGetHeight(glb.image);

  double delta0 = w*glb.value1;
  double delta1 = 0.5*w*glb.value2;

  u[0] = 0.0;
  u[1] = (double)w;
  u[2] = (double)w;
  u[3] = 0.0;

  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = (double)h;
  v[3] = (double)h;

  x[0] = -delta0;
  x[1] = (double)w+delta0;
  x[2] = (double)w-delta1;
  x[3] = delta1;

  y[0] = 0.0;
  y[1] = 0.0;
  y[2] = (double)h;
  y[3] = (double)h;

  homography4p(u, v, x, y, H);
  m3Inv(H, Hinv);

  if(glb.flags[3])
  {
   ogl_texture(Hinv);
   return;
  }

  glBegin(GL_POINTS);
  for (posy = 0; posy < h; posy++)
  {
   posxIni = ((((delta0+delta1)/h)*posy) - delta0);
   posxFin = (((-(delta0+delta1)/h)*posy) + delta0);
   for (posx = posxIni; posx < w+posxFin; posx++)
   {
    p1[0] = posx;
    p1[1] = posy;
    p1[2] = 1.0;
    m3MultAb(Hinv, p1, p2);
    p2[0] /= p2[2];
    p2[1] /= p2[2];

    /* space ellipse center (u0, v0) */
    u0 = p2[0];
    v0 = p2[1];

    if(glb.flags[0])
     imgGetPixel3fv(glb.image, (int)u0, (int)v0, rgb);
    else if(glb.flags[1])
     BilinearFilter(glb.image, rgb, u0, v0);
    else if(glb.flags[2])
    {
     /* ellipse vectors (ux, vx) and (uy, vy) */
     du[0] = 0.0;
     du[1] = 0.0;
     dv[0] = 0.0;
     dv[1] = 0.0;
     
     derivateMap(Hinv, posx, posy, du, dv);

     EccentricityCorr(du, dv);

     EWAFilter(glb.image, u0, v0, du[0], du[1], dv[0], dv[1], rgb);
    }
   
    glColor3fv(rgb);
    glVertex2i(posx,posy);
   }
  }
  glEnd();
 }
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int right_repaint_cb(Ihandle* self)
{
 int w = 256;
 int h = 256;
 double delta0 = 0.0;
 double delta1 = 0.0;

 IupGLMakeCurrent(self);        /* activates this GL Canvas as the current drawing area. */
 glClear(GL_COLOR_BUFFER_BIT);  /* clears the back buffer */

 my_texture( );

 if(glb.image)
 {
  w=imgGetWidth(glb.image);
  h=imgGetHeight(glb.image);
  delta0 = w*glb.value1;
  delta1 = 0.5*w*glb.value2;
 }

 glColor3f(1.f,1.f,0.f);
 glBegin(GL_LINE_LOOP);
 glVertex2d(-delta0,0.f);
 glVertex2d(w+delta0,0);
 glVertex2d(w-delta1,h);
 glVertex2d(delta1,h);
 glEnd();

 IupGLSwapBuffers(self);        /* swap data from back buffer to front buffer */
 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int resize_cb(Ihandle* self, int width, int height)
{
 int w = 256;
 int h = 256;
 double deltaW;
 double deltaH;

 if(glb.image)
 {
  w = imgGetWidth(glb.image);
  h = imgGetHeight(glb.image);
 }

 deltaW = (width-w)/2;
 deltaH = (height-h)/2;
 IupGLMakeCurrent(self);
 /* projection*/
 glViewport(0, 0, width, height);

 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 gluOrtho2D (-deltaW, (GLdouble)width-deltaW, -deltaH, (GLdouble)height-deltaH);

 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 return IUP_DEFAULT;
} 

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int idle_cb(void)
{
 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
void update_state ( )
{
 IupSetfAttribute(glb.status_label, "TITLE", "flags:%1d%1d%1d value1: %.2lf  value2: %.2lf", 
  glb.flags[0],glb.flags[1],glb.flags[2],
  glb.value1,glb.value2);
 left_repaint_cb(glb.left_canvas);
 right_repaint_cb(glb.right_canvas);
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int linear_cb(Ihandle *self)
{
 if (glb.flags[0] == 0)
  glb.flags[0] = 1;
 else
  glb.flags[0] = 0;
 
 update_state ( );
 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int bilinear_cb(Ihandle *self)
{
 if (glb.flags[1] == 0)
  glb.flags[1] = 1;
 else
  glb.flags[1] = 0;

 update_state ( );
 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int ewa_cb(Ihandle *self)
{
 if (glb.flags[2] == 0)
  glb.flags[2] = 1;
 else
  glb.flags[2] = 0;

 update_state ( );

 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int opengl_cb(Ihandle *self)
{
 if (glb.flags[3] == 0)
  glb.flags[3] = 1;
 else
  glb.flags[3] = 0;

 update_state ( );

 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int val1_cb(Ihandle *self, double value)
{
 glb.value1 = value;
 update_state ( );
 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int val2_cb(Ihandle *self, double value)
{
 if(value<=1.0)
 {
  glb.value2 = value;
  update_state ( );
  return IUP_DEFAULT;
 }
 update_state ( );
 return IUP_IGNORE;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
/* Ajusta o tamanho da janela do dialogo de acordo com o tamanho da imagem exibida */
static void update_dialog_size( )
{
#if 0
 if (glb.image)
 {
  int w,h;
  char buffer[512];
  w=imgGetWidth(glb.image);
  h=imgGetHeight(glb.image);
  sprintf(buffer,"%dx%d",w,h);
  IupSetAttribute(glb.left_canvas, IUP_RASTERSIZE, buffer);
  IupSetAttribute(glb.right_canvas, IUP_RASTERSIZE, buffer);
  IupSetAttribute(glb.dialog, IUP_RASTERSIZE, NULL);
  IupShowXY(glb.dialog, IUP_CENTER, IUP_CENTER);
  update_state ( );
 }
#endif
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int open_cb(Ihandle* self)
{
 int w,h,wtext,htext;
 float* pBuffer;
 int ret;
 char filename[1024] = ".\\*.*";

 /* if image already exists delete it */
 if(glb.image)
  imgDestroy(glb.image);

 /* get a file name */
 ret = IupGetFile(filename);
 if (ret == -1)
  return IUP_DEFAULT;

 glb.image = imgReadBMP(filename);
 if(glb.image==NULL)
  return IUP_DEFAULT;
 
 /* adjust for 2^n */
 w = imgGetWidth(glb.image);
 h = imgGetHeight(glb.image);
 wtext = (int) pow(2, ceil((log(w)/log(2))));
 htext = (int) pow(2, ceil((log(h)/log(2))));
 if ((wtext!=w)||(htext!=h))
 {
  Image* tmp=glb.image;
  glb.image = imgAdjust2eN(tmp);
  imgDestroy(tmp);
 }

 /* create texture */
 w = imgGetWidth(glb.image);
 h = imgGetHeight(glb.image);
 pBuffer = imgGetData(glb.image);
 glEnable(GL_TEXTURE_2D);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
 glTexImage2D(GL_TEXTURE_2D,0,3,w,h,0,GL_RGB,GL_FLOAT,pBuffer);
 glDisable(GL_TEXTURE_2D);

 left_repaint_cb(glb.left_canvas);
 right_repaint_cb(glb.right_canvas);
 update_dialog_size( );

 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int exit_cb(Ihandle *self)
{
 /* destroy buffers, disconnect devices, close files */
 if (glb.image) 
  imgDestroy(glb.image);
 
 /* free WTAB mem */
 freeMem();
 return IUP_CLOSE;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int about_cb(Ihandle *self)
{
 IupMessagef("About", "T2 -FCG - Filtros\n"
                      "Tecgraf/PUC-Rio\n"
                      " ---------------- \n"
                      "IUP Version %s\n"
                      " ---------------- \n"
                      "OpenGL:\n"
                      "  Vendor: %s\n"
                      "  Renderer: %s\n"
                      "  Version: %s\n"
                      , IUP_VERSION, 
                      glGetString(GL_VENDOR),glGetString(GL_RENDERER),glGetString(GL_VERSION));
 return IUP_DEFAULT;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
void mainMenuCreate( ) 
{
 Ihandle* file_menu = IupMenu(IupItem("Open", "open_cb"), IupItem("Exit", "exit_cb"), NULL);
 Ihandle* help_menu = IupMenu(IupItem("About...", "about_cb"), NULL);
 Ihandle* menu = IupMenu(IupSubmenu("File", file_menu), IupSubmenu("Help", help_menu), NULL);

 /* this will be used by the dialog */
 IupSetHandle("menu", menu);

 /* associate callbacks with names */
 IupSetFunction("exit_cb", (Icallback)exit_cb);
 IupSetFunction("about_cb", (Icallback)about_cb);
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
Ihandle* mainToolbarCreate( ) 
{
 Ihandle *toolbar, *filterFr, *val1, *val2, *open, *exit;

 /* creates the toolbar and its buttons */

 open = IupButton("", "open_cb");
 IupSetAttribute(open,"TIP","Opens image file.");
 IupSetAttribute(open,"IMAGE","IUP_IMGBUT_OPEN");
 IupSetFunction("open_cb", (Icallback)open_cb);

 exit = IupButton("exit", "exit_cb");
 IupSetAttribute(open,"TIP","Exits the program.");
 IupSetFunction("exit_cb", (Icallback)exit_cb);


 val2 = IupVal("HORIZONTAL");
 IupSetAttribute(val2, "MOUSEMOVE_CB", "user_val2_cb");
 IupSetFunction("user_val2_cb", (Icallback)val2_cb);

 val1 = IupVal("HORIZONTAL");
 IupSetAttribute(val1, "MOUSEMOVE_CB", "user_val1_cb");
 IupSetFunction("user_val1_cb", (Icallback)val1_cb);

 filterFr = IupFrame
            (
             IupRadio
             (
              IupHbox(
               IupToggle("Linear", "linear_cb"),  
               IupToggle("Bilinear", "bilinear_cb"),
               IupToggle("EWA", "ewa_cb"),
               IupToggle("OpenGL", "opengl_cb"),
              NULL)
             )
            );

 toolbar = IupHbox(
            open, 
            IupFill(),
            filterFr,
            IupFill(),
            val1,
            val2,
            IupFill(),
            exit,
           NULL);

 IupSetAttribute(filterFr, "TITLE", "Filter Type");

 IupSetAttribute(toolbar, "ALIGNMENT", "ACENTER");
 IupSetAttribute(toolbar, "MARGIN", "2x2");

 IupSetFunction("linear_cb",   (Icallback)linear_cb);  /* EDIT_HERE */
 IupSetFunction("bilinear_cb", (Icallback)bilinear_cb);
 IupSetFunction("ewa_cb",      (Icallback)ewa_cb);
 IupSetFunction("opengl_cb",   (Icallback)opengl_cb);

 return toolbar;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
void createMainDialog(void)
{
 Ihandle *dialog, *statusbar, *toolbar, *box,
         *label, *left_canvas, *right_canvas;


 /* initialize interface */
 toolbar = mainToolbarCreate( );

 left_canvas = IupGLCanvas("left_repaint_cb");
 IupSetAttribute(left_canvas, "BORDER", "NO");
 IupSetAttribute(left_canvas, "BUFFER", "DOUBLE");   /* use double buffer */
 IupSetAttribute(left_canvas, "RESIZE_CB", "resize_cb");   /* configure the resize callback */
 IupSetAttribute(left_canvas, "RASTERSIZE", "384x384"); /* initial size */
 glb.left_canvas = left_canvas;

 right_canvas = IupGLCanvas("right_repaint_cb");
 IupSetAttribute(right_canvas, "BORDER", "NO");
 IupSetAttribute(right_canvas, "BUFFER", "DOUBLE");  /* use double buffer */
 IupSetAttribute(right_canvas, "RESIZE_CB", "resize_cb");  /* configure the resize callback */
 IupSetAttribute(right_canvas, "RASTERSIZE", "384x384"); /* initial size */
 glb.right_canvas = right_canvas;

 IupSetFunction("resize_cb", (Icallback)resize_cb);
 IupSetFunction("left_repaint_cb", (Icallback)left_repaint_cb);
 IupSetFunction("right_repaint_cb", (Icallback)right_repaint_cb);

 /* the status bar is just a label to put some usefull information in run time */
 label = IupLabel("status");
 IupSetAttribute(label, "EXPAND", "HORIZONTAL");
 IupSetAttribute(label, "FONT", "COURIER_NORMAL_10");
 statusbar = IupSetAttributes(IupHbox(
               IupFrame(IupHbox(label, NULL)), 
               NULL), "MARGIN=5x5");
 glb.status_label = label;

 /* this is the most external box that puts together
    the toolbar, the two canvas and the status bar */
 box = IupVbox(
        toolbar,
        IupSetAttributes(IupHbox(
         left_canvas, 
         right_canvas, 
         NULL),
        "GAP=5"),
        statusbar, 
       NULL);

 /* create the dialog and set its attributes */

 mainMenuCreate( );

 dialog = IupDialog(box);
 IupSetAttribute(dialog, "MENU", "menu");     /* configure the menu */
 IupSetAttribute(dialog, "CLOSE_CB", "exit_cb");
 IupSetAttribute(dialog, "TITLE", "IUPGLCapture");
 IupSetAttribute(dialog, "RASTERSIZE", NULL); /* initial size */
 IupSetHandle("dialog", dialog);
 IupSetGlobal(IUP_PARENTDIALOG, "dialog");     /* to be used by the idle callback */

 glb.dialog = dialog;

 IupShowXY(dialog, IUP_CENTER, IUP_CENTER);
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
int main(int argc, char* argv[]) 
{
 /* IUP initialization */
 IupOpen( );
 IupControlsOpen( );
 IupGLCanvasOpen( );
 IupImageLibOpen( );

 /* Create and show the main dialog */
 createMainDialog();

 glClearColor(0.0, 0.0, 0.0, 1.0);

 /* WTAB initialization */
 initWTAB();

 /* IUP event loop */
 IupMainLoop();

 /* IUP closing */
 IupControlsClose();
 IupImageLibClose( );
 IupClose();

 return 0;
}
