/* IUP/OpenGL image example

  Uses IUP for user interface
       IM for image I/O and capture
       OpenGL for drawing

  Needs IUP: comctl32.lib iup.lib iupcontrols.lib iupgl.lib cd.lib cdiup.lib 
        OpenGL: opengl32.lib glu32.lib

  Libraries Available at:
    http://www.tecgraf.puc-rio.br/iup/

  There are two canvases. 
  One for that uses the OpenGL process (left) and 
  one for you to disply your results (right). 
  
  Version:
            0.9 

  Authors:
            Marcelo Gattass
            Antonio Scuri

  Tecgraf/PUC-Rio - 2007
*/

#ifdef WIN32            
#include <windows.h>    /* necessary because of the Microsoft OpenGL headers dependency */
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <iup.h>
#include <iupcontrols.h>
#include <iupgl.h>

#include "image.h"

struct _appData
{
    /* Global to help enable/disable items in the interface */
    Image*  image;
    Ihandle *left_canvas, *right_canvas, *status_label;
    int flags[3];
    double value1, value2;
} app_Data;



/*
 * Callbacks
 */

int flag1_cb(Ihandle *self)
{
  if (app_data->flags[0] == 0)
    app_data->flags[0] = 1;
  else
    app_data->flags[0] = 0;

  IupSetfAttribute(app_data->status_label, "TITLE", "flags:%1d%1d%1d value1: %.2lf  value2: %.2lf", 
      app_data->flags[0],app_data->flags[1],app_data->flags[2],
      app_data->value1,app_data->value2);

  return IUP_DEFAULT;
}

int flag2_cb(Ihandle *self)
{
  if (app_data->flags[1] == 0)
    app_data->flags[1] = 1;
  else
    app_data->flags[1] = 0;

  IupSetfAttribute(app_data->status_label, "TITLE", "flags:%1d%1d%1d value1: %.2lf  value2: %.2lf", 
      app_data->flags[0],app_data->flags[1],app_data->flags[2],
      app_data->value1,app_data->value2);

  return IUP_DEFAULT;
}

int flag3_cb(Ihandle *self)
{
  if (app_data->flags[2] == 0)
    app_data->flags[2] = 1;
  else
    app_data->flags[2] = 0;

  IupSetfAttribute(app_data->status_label, "TITLE", "flags:%1d%1d%1d value1: %.2lf  value2: %.2lf", 
      app_data->flags[0],app_data->flags[1],app_data->flags[2],
      app_data->value1,app_data->value2);

  return IUP_DEFAULT;
}


int val1_cb(Ihandle *self, double value)
{
   app_data->value1 = value;

   IupSetfAttribute(app_data->status_label, "TITLE", "flags:%1d%1d%1d value1: %.2lf  value2: %.2lf", 
       app_data->flags[0],app_data->flags[1],app_data->flags[2],
       app_data->value1,app_data->value2);
   return IUP_DEFAULT;
}

int val2_cb(Ihandle *self, double value)
{
    app_data->value2 = value;
    IupSetfAttribute(app_data->status_label, "TITLE", "flags:%1d%1d%1d value1: %.2lf  value2: %.2lf", 
        app_data->flags[0],app_data->flags[1],app_data->flags[2],
        app_data->value1,app_data->value2);
    return IUP_DEFAULT;
}

int repaint(Ihandle* self, int left)
{
  IupGLMakeCurrent(self);        /* activates this GL Canvas as the current drawing area. */
  glClear(GL_COLOR_BUFFER_BIT);  /* clears the back buffer */

  if (app_data->image!=NULL)
  {
    int w = imgGetWidth(app_data->image);
    int h = imgGetHeight(app_data->image);
    float* pBuffer = imgGetData(app_data->image);

    /* Draws the captured image at (0,0) */
    glRasterPos2f(0.f, 0.f);

    if (left)
      glDrawPixels(w, h, GL_RGB, GL_FLOAT, pBuffer);
    else
      glDrawPixels(w, h, GL_RGB, GL_FLOAT, pBuffer);
  }

  IupGLSwapBuffers(self);        /* swap data from back buffer to front buffer */
  return IUP_DEFAULT;
} 

int left_repaint_cb(Ihandle* self)
{
  return repaint(self, 1);
} 

int right_repaint_cb(Ihandle* self)
{
  return repaint(self, 0);
} 

void OpenGLInit(int width, int height)
{
  glClearColor(0., 0., 0., 1.0);                   /* window background */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);           /* data alignment is 1 */

  glViewport(0, 0, width, height);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D (0.0, (GLdouble)width, 0.0, (GLdouble)height);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

int resize_cb(Ihandle* self, int w, int h)
{
  IupGLMakeCurrent(self);
  OpenGLInit(w, h);
  return IUP_DEFAULT;
} 


int idle_cb(void)
{
  return IUP_DEFAULT;
}

int open_cb(Ihandle* self)
{
  int ret;
  char filename[1024] = ".\\*.*";

  /* if image already exists delete it */
  if (app_data->image)
      imgDestroy(app_data->image);

  /* get a file name */
  ret = IupGetFile(filename);
  if (ret == -1)
    return IUP_DEFAULT;

  app_data->image = imgReadBMP(filename);

  return IUP_DEFAULT;
}



int exit_cb(Ihandle *self)
{

  /* destroy buffers, disconnect devices, close files */

  if (app_data->image) 
      imgDestroy(app_data->image);

   free(app_data);

  return IUP_CLOSE;
}

int about_cb(Ihandle *self)
{
  IupMessagef("About", "T2 -FCG - Filtros\n"
                       "Tecgraf/PUC-Rio\n"
                       " ---------------- \n"
                       "IUP Version %s\n"
                       " ---------------- \n"
                       "OpenGL:\n"
                       "  Vendor: %s\n"
                       "  Renderer: %s\n"
                       "  Version: %s\n"
                       , IUP_VERSION, 
                         glGetString(GL_VENDOR),glGetString(GL_RENDERER),glGetString(GL_VERSION));
  return IUP_DEFAULT;
}

void mainMenuCreate(appData* app_data) 
{
  Ihandle* file_menu = IupMenu(
      IupItem( "Open", "open_cb"),
      IupItem( "Exit", "exit_cb"),
     NULL
  );


  Ihandle* help_menu = IupMenu(
     IupItem( "About...", "about_cb"),
     NULL
  );

  Ihandle* menu = IupMenu(
     IupSubmenu("File", file_menu),
     IupSubmenu("Help", help_menu),
     NULL
  );


  /* this will be used by the dialog */
  IupSetHandle("menu", menu);

  /* associate callbacks with names */
  IupSetFunction("exit_cb", (Icallback)exit_cb);
  IupSetFunction("about_cb", (Icallback)about_cb);
}

Ihandle* mainToolbarCreate(appData* app_data) 
{
  Ihandle *toolbar, *val1, *val2, *open, *exit;

  /* creates the toolbar and its buttons */

  open = IupButton("", "open_cb");
  IupSetAttribute(open,"TIP","Opens image file.");
  IupSetAttribute(open,"IMAGE","IUP_IMGBUT_OPEN");
  IupSetFunction("open_cb", (Icallback)open_cb);

  exit = IupButton("exit", "exit_cb");
  IupSetAttribute(open,"TIP","Exits the program.");
  IupSetFunction("exit_cb", (Icallback)exit_cb);


  val2 = IupVal("HORIZONTAL");
  IupSetAttribute(val2, "MOUSEMOVE_CB", "user_val2_cb");
  IupSetFunction("user_val2_cb", (Icallback)val2_cb);

  val1 = IupVal("HORIZONTAL");
  IupSetAttribute(val1, "MOUSEMOVE_CB", "user_val1_cb");
  IupSetFunction("user_val1_cb", (Icallback)val1_cb);

  toolbar = IupHbox(
       open, 
       IupFill(),
       IupToggle("flag1", "flag1_cb"),  
       IupToggle("flag2", "flag2_cb"),
       IupToggle("flag3", "flag3_cb"),
       val1,
       val2,
       IupFill(),
       exit,
       NULL);

  IupSetAttribute(toolbar, "ALIGNMENT", "ACENTER");
  IupSetAttribute(toolbar, "MARGIN", "2x2");

  IupSetFunction("flag1_cb", (Icallback)flag1_cb);  /* EDIT_HERE */
  IupSetFunction("flag2_cb", (Icallback)flag2_cb);
  IupSetFunction("flag3_cb", (Icallback)flag3_cb);

  return toolbar;
}

void createMainDialog(void)
{
  appData* app_data;
  Ihandle *dialog, *statusbar, *toolbar, *box,
          *label, *left_canvas, *right_canvas;

  /* initialize application context */
  app_data = malloc(sizeof(appData));
  memset(app_data, 0, sizeof(appData));

  /* initialize interface */
  toolbar = mainToolbarCreate(app_data);

  left_canvas = IupGLCanvas("left_repaint_cb");
  IupSetAttribute(left_canvas, "BORDER", "NO");
  IupSetAttribute(left_canvas, "BUFFER", "DOUBLE");   /* use double buffer */
  IupSetAttribute(left_canvas, "RESIZE_CB", "resize_cb");   /* configure the resize callback */
  app_data->left_canvas = left_canvas;

  right_canvas = IupGLCanvas("right_repaint_cb");
  IupSetAttribute(right_canvas, "BORDER", "NO");
  IupSetAttribute(right_canvas, "BUFFER", "DOUBLE");  /* use double buffer */
  IupSetAttribute(right_canvas, "RESIZE_CB", "resize_cb");  /* configure the resize callback */
  app_data->right_canvas = right_canvas;

  IupSetFunction("resize_cb", (Icallback)resize_cb);
  IupSetFunction("left_repaint_cb", (Icallback)left_repaint_cb);
  IupSetFunction("right_repaint_cb", (Icallback)right_repaint_cb);

  /* the status bar is just a label to put some usefull information in run time */
  label = IupLabel("status");
  IupSetAttribute(label, "EXPAND", "HORIZONTAL");
  IupSetAttribute(label, "FONT", "COURIER_NORMAL_10");
  statusbar = IupSetAttributes(IupHbox(
                IupFrame(IupHbox(label, NULL)), 
                NULL), "MARGIN=5x5");
  app_data->status_label = label;

  /* this is the most external box that puts together
     the toolbar, the two canvas and the status bar */
  box = IupVbox(
          toolbar,
          IupSetAttributes(IupHbox(
            left_canvas, 
            right_canvas, 
            NULL), "GAP=5"),
          statusbar, 
          NULL);

  /* create the dialog and set its attributes */

  mainMenuCreate(app_data);

  dialog = IupDialog(box);
  IupSetAttribute(dialog, "MENU", "menu");     /* configure the menu */
  IupSetAttribute(dialog, "CLOSE_CB", "exit_cb");
  IupSetAttribute(dialog, "TITLE", "IUPGLCapture");
  IupSetAttribute(dialog, "RASTERSIZE", "680x480"); /* initial size */
  IupSetAttribute(dialog, "SHRINK", "YES");
  IupSetHandle("dialog", dialog);
  IupSetGlobal(IUP_PARENTDIALOG, "dialog");     /* to be used by the idle callback */

  IupSetAttribute(dialog, "DATA", (char*)app_data);
  IupSetGlobal("DATA", (char*)app_data);

  IupShowXY(dialog, IUP_CENTER, IUP_CENTER);
}

int main(int argc, char* argv[]) 
{
  /* IUP initialization */
  IupOpen( );
  IupControlsOpen( );
  IupGLCanvasOpen( );
  IupImageLibOpen( );

  /* Create and show the main dialog */
  createMainDialog();

     /* IUP event loop */
     IupMainLoop();

  /* IUP closing */
  IupControlsClose();
  IupImageLibClose( );
  IupClose();

  return 0;
}
